from numba import cuda, jit, float32
import numpy as np
import math

from descriptools.helpers import divisor


def downslope_sequential(dem,
                         flow_direction,
                         px,
                         elevation_difference,
                         downslope=np.array([])):
    '''
    Downslope sequential method. Also responsible for fixing cells that could
    not be simulated in the gpu.

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    flow_direction : int
        Flow direction.
    px : int or float
        Raster pixel dimension.
    elevation_difference : int
        Elevation difference.
    downslope : Downslope float array, optional
        Downslope array. The default is np.array([]).

    Returns
    -------
    downslope : float array
        Downslope index array.

    '''
    new = 0
    if downslope.size == 0:
        downslope = np.zeros([len(dem), len(dem[0])], dtype='float32')
        new = 1

    for i in range(622, len(downslope), 1):
        for j in range(2020, len(downslope[0]), 1):
            if dem[i, j] == -100:
                downslope[i, j] = -100
                continue
            elif new == 0 and downslope[i, j] != -50:
                continue
            else:
                y = i
                x = j
                dist = 0
                loop = 0
                is_nan = 0
                while (dem[i, j] - dem[y, x] < elevation_difference):
                    if y == 0 and (flow_direction[y, x] == 32
                                   or flow_direction[y, x] == 64
                                   or flow_direction[y, x] == 128):
                        is_nan = 1
                        break
                    elif y == len(downslope) - 1 and (
                            flow_direction[y, x] == 2 or flow_direction[y, x]
                            == 4 or flow_direction[y, x] == 8):
                        is_nan = 1
                        break
                    elif x == 0 and (flow_direction[y, x] == 32
                                     or flow_direction[y, x] == 16
                                     or flow_direction[y, x] == 8):
                        is_nan = 1
                        break
                    elif x == len(downslope[0]) - 1 and (
                            flow_direction[y, x] == 128 or flow_direction[y, x]
                            == 1 or flow_direction[y, x] == 2):
                        is_nan = 1
                        break

                    if flow_direction[y, x] == 1:
                        if dem[y, x + 1] == -100:
                            is_nan = 1
                            break
                        x += 1
                        dist += px
                    elif flow_direction[y, x] == 2:
                        if dem[y + 1, x + 1] == -100:
                            is_nan = 3
                            break
                        x += 1
                        y += 1
                        dist += px * math.sqrt(2.0)  #Arrumar raiz aqui
                    elif flow_direction[y, x] == 4:
                        if dem[y + 1, x] == -100:
                            is_nan = 1
                            break
                        y += 1
                        dist += px
                    elif flow_direction[y, x] == 8:
                        if dem[y + 1, x - 1] == -100:
                            is_nan = 1
                            break
                        x -= 1
                        y += 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == 16:
                        if dem[y, x - 1] == -100:
                            is_nan = 1
                            break
                        x -= 1
                        dist += px
                    elif flow_direction[y, x] == 32:
                        if dem[y - 1, x - 1] == -100:
                            is_nan = 1
                            break
                        x -= 1
                        y -= 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == 64:
                        if dem[y - 1, x] == -100:
                            is_nan = 1
                            break
                        y -= 1
                        dist += px
                    elif flow_direction[y, x] == 128:
                        if dem[y - 1, x + 1] == -100:
                            is_nan = 1
                            break
                        x += 1
                        y -= 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == -100:
                        is_nan = 1
                        break

                    if y >= len(downslope):
                        y -= 1
                        if x >= len(downslope[0]):
                            x -= 1
                        break
                    elif x >= len(downslope[0]):
                        x -= 1
                        if y >= len(downslope):
                            y -= 1
                        break

                    if dem[y, x] == -100:
                        is_nan = 1
                        break

                    loop += 1

                    if loop == 5000:
                        break

                if is_nan > 1:
                    downslope[i, j] = -100
                else:
                    downslope[i, j] = (dem[i, j] - dem[y, x]) / dist

    return downslope


@jit
def downslope_sequential_jit(dem,
                             flow_direction,
                             px,
                             elevation_difference,
                             downslope=np.array([[], []], 'float32')):
    '''
    Downslope sequential method. Also responsible for fixing cells that could
    not be simulated in the gpu.

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    flow_direction : int
        Flow direction.
    px : int or float
        Raster pixel dimension.
    elevation_difference : int
        Elevation difference.
    downslope : Downslope float array, optional
        Downslope array. The default is np.array([]).

    Returns
    -------
    downslope : float array
        Downslope index array.

    '''
    new = 0
    if downslope.size == 0:
        downslope = np.zeros(dem.shape, dtype=float32)
        new = 1

    for i in range(0, len(downslope), 1):
        for j in range(0, len(downslope[0]), 1):
          
            if dem[i, j] == -100:
                downslope[i, j] = -100
                continue
            elif new == 0 and downslope[i, j] != -50:
                continue
            else:
                y = i
                x = j
                dist = 0
                loop = 0
                is_nan = 0
                while (dem[i, j] - dem[y, x] < elevation_difference):
                    if y == 0 and (flow_direction[y, x] == 32
                                   or flow_direction[y, x] == 64
                                   or flow_direction[y, x] == 128):
                        is_nan = 1
                        break
                    elif y == len(downslope) - 1 and (
                            flow_direction[y, x] == 2 or flow_direction[y, x]
                            == 4 or flow_direction[y, x] == 8):
                        is_nan = 1
                        break
                    elif x == 0 and (flow_direction[y, x] == 32
                                     or flow_direction[y, x] == 16
                                     or flow_direction[y, x] == 8):
                        is_nan = 1
                        break
                    elif x == len(downslope[0]) - 1 and (
                            flow_direction[y, x] == 128 or flow_direction[y, x]
                            == 1 or flow_direction[y, x] == 2):
                        is_nan = 1
                        break

                    if flow_direction[y, x] == 1:
                        if dem[y, x + 1] == -100:
                            is_nan = 1
                            break
                        x += 1
                        dist += px
                    elif flow_direction[y, x] == 2:
                        if dem[y + 1, x + 1] == -100:
                            is_nan = 1
                            break
                        x += 1
                        y += 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == 4:
                        if dem[y + 1, x] == -100:
                            is_nan = 1
                            break
                        y += 1
                        dist += px
                    elif flow_direction[y, x] == 8:
                        if dem[y + 1, x - 1] == -100:
                            is_nan = 1
                            break
                        x -= 1
                        y += 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == 16:
                        if dem[y, x - 1] == -100:
                            is_nan = 1
                            break
                        x -= 1
                        dist += px
                    elif flow_direction[y, x] == 32:
                        if dem[y - 1, x - 1] == -100:
                            is_nan = 1
                            break
                        x -= 1
                        y -= 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == 64:
                        if dem[y - 1, x] == -100:
                            is_nan = 1
                            break
                        y -= 1
                        dist += px
                    elif flow_direction[y, x] == 128:
                        if dem[y - 1, x + 1] == -100:
                            is_nan = 1
                            break
                        x += 1
                        y -= 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == -100:
                        is_nan = 1
                        break

                    if y >= len(downslope):
                        y -= 1
                        if x >= len(downslope[0]):
                            x -= 1
                        break
                    elif x >= len(downslope[0]):
                        x -= 1
                        if y >= len(downslope):
                            y -= 1
                        break

                    if dem[y, x] == -100:
                        is_nan = 1
                        break

                    loop += 1

                    if loop == 5000:
                        break

                if is_nan == 1:
                    if dist == 0:
                        downslope[i, j] = 0
                    else:
                        downslope[i, j] = (dem[i, j] - dem[y, x]) / dist
                else:
                    downslope[i, j] = (dem[i, j] - dem[y, x]) / dist

    return downslope


def downsloper(dem,
               flow_direction,
               px,
               elevation_difference,
               column_division=0,
               row_division=0):
    '''
    Method responsible for the partioning of the matrix for the downslope.

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    flow_direction : int
        Flow direction.
    px : int or float
        Raster pixel dimension.
    elevation_difference : int
        Elevation difference.
    column_division : int, optional
        Number of vertical divisions. The default is 0.
    row_division : int, optional
        Number of horizontal divisions. The default is 0.

    Returns
    -------
    downslope : float array
        Downslope index array.

    '''
    row_size = len(dem)
    col_size = len(dem[0])

    bRow, bCol = divisor(row_size, col_size, row_division, column_division)

    downslope = np.zeros((row_size, col_size), dtype='float32')

    bRow = np.insert(bRow, row_division, row_size)
    bRow = np.insert(bRow, 0, 0)
    bCol = np.insert(bCol, column_division, col_size)
    bCol = np.insert(bCol, 0, 0)

    for m in range(0, row_division + 1, 1):
        for n in range(0, column_division + 1, 1):
            mS = bRow[m]
            mE = bRow[m + 1]
            nS = bCol[n]
            nE = bCol[n + 1]

            downslope[mS:mE,
                      nS:nE] = downslope_cpu(dem[mS:mE, nS:nE],
                                             flow_direction[mS:mE, nS:nE], px,
                                             elevation_difference)
                                             

                                             
    downslope = downslope_sequential_jit(dem, flow_direction, px,
                                         elevation_difference, downslope)

    return downslope


def downslope_cpu(dem,
                  flow_direction,
                  px,
                  elevation_difference,
                  blocks=0,
                  threads=0):
    '''
    Method responsible for the host/device data transfer

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    flow_direction : int
        Flow direction.
    px : int or float
        Raster pixel dimension.
    elevation_difference : int
        Elevation difference.
    blocks : int, optional
        Number of block of threads. The default is 0.
    threads : int, optional
        number of threads in each block. The default is 0.

    Returns
    -------
    downslope : float array
        Downslope index array.

    '''
    row = len(dem)
    col = len(dem[0])
    
    if blocks == 0 and threads == 0:
        threads = 256
        blocks = math.ceil((row * col) / threads)
        # FIX: this!!!!
    dem = np.asarray(dem).reshape(-1)
    flow_direction = np.asarray(flow_direction).reshape(-1)
    downslope = np.zeros((row * col))

    dem = cuda.to_device(dem)
    flow_direction = cuda.to_device(flow_direction)
    downslope = cuda.to_device(downslope)

    downslope_gpu[blocks, threads](dem, flow_direction, downslope, px,
                                   elevation_difference, col, row)
    
    downslope = downslope.copy_to_host()

    downslope = downslope.reshape(row, col)  

    return downslope


@cuda.jit
def downslope_gpu(dem, flow_direction, downslope, px, elevation_difference,
                  col, row):
    '''
    GPU Downslope index method

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    flow_direction : int
        Flow direction.
    downslope : float array
        Downslope index array.
    px : int or float
        Raster pixel dimension.
    elevation_difference : int
        Elevation difference.
    row : int
        Number of rows of the 2-D matrix.
    col : int
        Number of rows of the 2-D matrix.

    '''
    i = cuda.grid(1)
    if i >= 0 and i < row * col:
        if dem[i] <= -100:
            downslope[i] = -100
        else:
            pos = i
            is_nan = 0
            dist = 0
            loop = 0
            out = 0
            while (dem[i] - dem[pos] < elevation_difference):
                if pos < col and (flow_direction[pos] == 32
                                  or flow_direction[pos] == 64
                                  or flow_direction[pos] == 128):
                    out = 1
                    break
                elif pos % col == 0 and (flow_direction[pos] == 8
                                         or flow_direction[pos] == 16
                                         or flow_direction[pos] == 32):
                    out = 1
                    break
                elif pos % col == (col - 1) and (flow_direction[pos] == 128
                                                 or flow_direction[pos] == 1
                                                 or flow_direction[pos] == 2):
                    out = 1
                    break
                elif pos >= (row - 1) * col and (flow_direction[pos] == 2
                                                 or flow_direction[pos] == 4
                                                 or flow_direction[pos] == 8):
                    out = 1
                    break

                if flow_direction[pos] == 1:
                    pos += 1
                    dist += px
                elif flow_direction[pos] == 2:
                    pos += 1 + col
                    dist += (px * math.sqrt(2.0))
                elif flow_direction[pos] == 4:
                    pos += col
                    dist += px
                elif flow_direction[pos] == 8:
                    pos += col - 1
                    dist += (px * math.sqrt(2.0))
                elif flow_direction[pos] == 16:
                    pos += -1
                    dist += px
                elif flow_direction[pos] == 32:
                    pos += -1 - col
                    dist += (px * math.sqrt(2.0))
                elif flow_direction[pos] == 64:
                    pos += -col
                    dist += px
                elif flow_direction[pos] == 128:
                    pos += -col + 1
                    dist += (px * math.sqrt(2.0))
                if flow_direction[pos] == -100:
                    is_nan = 1
                    break

                loop += 1
                if loop == 5000:
                    is_nan = 1
                    break

                if dem[pos] == -100:
                    is_nan = 1
                    break
            if is_nan == 1:
                downslope[i] = -50
            elif out == 1:
                downslope[i] = -50

            else:
                downslope[i] = (dem[i] - dem[pos]) / dist


# -*- coding: utf-8 -*-
from numba import cuda, jit, float32
import numpy as np
import math

from descriptools.helpers import divisor


def topographic_index_sequential(flow_accumulation, slope, px):
    '''
    Sequential method of the Topographic index calculation

    Parameters
    ---------- 
    flow_accumulation : int
        Flow accumulation. Servers as the...
    slope : float
        Highest slope to a neighbouring cell. Surrogate to...
    px : int or float
        Raster dimension size.

    Returns
    -------
    topographic_index : float
        Topographic index.

    '''
    topographic_index = np.where(
        slope == -100, -100,
        np.log((np.where(flow_accumulation == 0, 1, flow_accumulation) *
                np.power(px, 2)) / (np.tan(slope))))

    return topographic_index


@jit
def topographic_index_sequential_jit(flow_accumulation, slope, px):
    topographic_index = np.zeros(slope.shape, dtype=float32)

    for i in range(0, len(slope), 1):
        for j in range(0, len(slope[0]), 1):
            if slope[i, j] == -100:
                topographic_index[i, j] = -100
            else:
                if flow_accumulation[i, j] == 0:
                    topographic_index[i, j] = np.log(
                        (1 * px * px) / (np.tan(slope[i, j]) + 0.01))
                else:
                    topographic_index[i, j] = np.log(
                        (flow_accumulation[i, j] * px * px) /
                        (np.tan(slope[i, j]) + 0.01))

    return topographic_index


@jit
def modified_topographic_index_sequential_jit(flow_accumulation, slope, px,
                                              expoent):
    modified_topographic_index = np.zeros(slope.shape, dtype=float32)

    for i in range(0, len(slope), 1):
        for j in range(0, len(slope[0]), 1):
            if slope[i, j] == -100:
                modified_topographic_index[i, j] = -100
            else:
                if flow_accumulation[i, j] == 0:
                    modified_topographic_index[i, j] = np.log(
                        np.power(1 * px * px, expoent) /
                        (np.tan(slope[i, j]) + 0.01))
                else:
                    modified_topographic_index[i, j] = np.log(
                        np.power(flow_accumulation[i, j] * px * px, expoent) /
                        (np.tan(slope[i, j]) + 0.01))

    return modified_topographic_index


def modified_topographic_index_sequential(flow_accumulation, slope, px,
                                          expoent):
    '''
    Sequential method of the Modified Topographic Index calculation

    Parameters
    ----------
    flow_accumulation : int
        Flow accumulation. Servers as the...
    slope : float
        Highest slope to a neighbouring cell. Surrogate to...
    px : int or float
        Raster dimension size.
    expoent : float
        Expoent (<1) calibrated for the area

    Returns
    -------
    modified_topographic_index : float
        Modified Topographic index.
    '''

    modified_topographic_index = np.where(
        slope == -100, -100,
        np.log(
            np.power((np.where(flow_accumulation == 0, 1, flow_accumulation) *
                      np.power(px, 2)), expoent) / (np.tan(slope))))

    return modified_topographic_index


def topographic_index(flow_accumulation,
                      slope,
                      px,
                      n_top,
                      div_col=0,
                      div_row=0):
    '''
    Method responsible for the partioning of the matrix

    Parameters
    ----------
    flow_accumulation : int
        Flow accumulation. Servers as the...
    slope : float
        Highest slope to a neighbouring cell. Surrogate to...
    px : int or float
        Raster dimension size.
    n_top : float
        Expoent (<1) calibrated for the area
    div_col : int, optional
        Number of vertical divisions. The default is 0.
    div_row : int, optional
        Number of horizontal divisions. The default is 0.

    Returns
    -------
    topographic_index : float
        Topographic index..
    modified_topographic_index : float
        Modified Topographic index.

    '''
    row_size = len(flow_accumulation)
    col_size = len(flow_accumulation[0])

    bRow, bCol = divisor(row_size, col_size, div_row, div_col)

    topographic_index = np.zeros((row_size, col_size))
    modified_topographic_index = np.zeros((row_size, col_size))

    bRow = np.insert(bRow, div_row, row_size)
    bRow = np.insert(bRow, 0, 0)
    bCol = np.insert(bCol, div_col, col_size)
    bCol = np.insert(bCol, 0, 0)

    for m in range(0, div_row + 1, 1):
        for n in range(0, div_col + 1, 1):

            mS = bRow[m]
            mE = bRow[m + 1]
            nS = bCol[n]
            nE = bCol[n + 1]

            topographic_index[mS:mE, nS:nE], modified_topographic_index[
                mS:mE,
                nS:nE] = topographic_index_cpu(flow_accumulation[mS:mE, nS:nE],
                                               slope[mS:mE, nS:nE], px, n_top)

    return topographic_index, modified_topographic_index


def topographic_index_cpu(flow_accumulation,
                          slope,
                          px,
                          expoent,
                          blocks=0,
                          threads=0):
    '''
    Method responsible for the host/device data transfer

    Parameters
    ----------
    flow_accumulation : int
        Flow accumulation. Servers as the...
    slope : float
        Highest slope to a neighbouring cell. Surrogate to...
    px : int or float
        Raster dimension size.
    expoent : float
        Expoent (<1) calibrated for the area
    blocks : int, optional
        Number of block of threads. The default is 0.
    threads : int, optional
        number of threads in each block. The default is 0.

    Returns
    -------
    topographic_index : float
        Topographic index.
    modified_topographic_index : float
        Modified Topographic index.

    '''
    row = len(flow_accumulation)
    col = len(flow_accumulation[0])
    if blocks == 0 and threads == 0:
        threads = 256
        blocks = math.ceil((row * col) / threads)

    flow_accumulation = np.asarray(flow_accumulation).reshape(-1)
    slope = np.asarray(slope).reshape(-1)
    topographic_index = np.zeros((row * col), dtype='float32')
    modified_topographic_index = np.zeros((row * col), dtype='float32')

    flow_accumulation = cuda.to_device(flow_accumulation)
    slope = cuda.to_device(slope)
    topographic_index = cuda.to_device(topographic_index)
    modified_topographic_index = cuda.to_device(modified_topographic_index)

    topographic_index_gpu[blocks, threads](flow_accumulation, slope,
                                           topographic_index, px)
    modified_topographic_index_gpu[blocks, threads](flow_accumulation, slope,
                                                    modified_topographic_index,
                                                    px, expoent)

    topographic_index = topographic_index.copy_to_host()
    modified_topographic_index = modified_topographic_index.copy_to_host()

    topographic_index = topographic_index.reshape(row, col)
    modified_topographic_index = modified_topographic_index.reshape(row, col)

    return topographic_index, modified_topographic_index


@cuda.jit
def topographic_index_gpu(flow_accumulation, slope, topographic_index, px):
    '''
    GPU Topographic index method

    Parameters
    ----------
    flow_accumulation : int
        Flow accumulation. Servers as the...
    slope : float
        Highest slope to a neighbouring cell. Surrogate to...
    topographic_index : float
        Topographic index.
    px : int or float
        Raster dimension size.

    '''
    i = cuda.grid(1)
    if i >= 0 and i < len(slope):
        if flow_accumulation[i] <= -100:
            topographic_index[i] = -100
        else:
            if flow_accumulation[i] == 0:
                topographic_index[i] = math.log(
                    (1 * px**2) / (math.tan(slope[i] + 0.01)))
            else:
                topographic_index[i] = math.log(
                    (flow_accumulation[i] * px**2) /
                    (math.tan(slope[i] + 0.01)))


@cuda.jit
def modified_topographic_index_gpu(flow_accumulation, slope,
                                   modified_topographic_index, px, expoent):
    '''
    GPU Modified Topographic index method

    Parameters
    ----------
    flow_accumulation : int
        Flow accumulation. Servers as the...
    slope : float
        Highest slope to a neighbouring cell. Surrogate to...
    modified_topographic_index : float
        Modified Topographic index.
    px : int or float
        Raster dimension size.
    expoent : float
        Expoent (<1) calibrated for the area

    '''
    i = cuda.grid(1)
    if i >= 0 and i < len(slope):
        if flow_accumulation[i] <= -100:
            modified_topographic_index[i] = -100
        else:
            if flow_accumulation[i] == 0:
                modified_topographic_index[i] = math.log(
                    ((1 * px**2)**expoent) / (math.tan(slope[i] + 0.01)))
            else:
                modified_topographic_index[i] = math.log(
                    ((flow_accumulation[i] * px**2)**expoent) /
                    (math.tan(slope[i] + 0.01)))



from numba import cuda, jit, float32
import numpy as np
import math

from descriptools.helpers import divisor


def geomorphic_flood_index_sequential(hand, flow_accumulation, indices,
                                      expoent, scale_factor):
    '''
    Sequential method for the GFI index

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    flow_accumulation : int
        Flow accumulation. Represent the river as source of hazard.
    indices : int array
        River cell index.
    expoent : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.

    Returns
    -------
    geomorphic_flood_index : float array
        geomorphic flood index array.

    '''
    river_flow_accumulation = river_accumulation(flow_accumulation, indices)

    geomorphic_flood_index = np.where(
        hand == -100, -100,
        np.where(
            hand == 0, 0,
            np.log(scale_factor * (np.power(
                np.where(river_flow_accumulation == 0, 1,
                         river_flow_accumulation), expoent)) / hand)))

    return geomorphic_flood_index


@jit
def geomorphic_flood_index_sequential_jit(hand, flow_accumulation, indices,
                                          expoent, scale_factor):
    river_flow_accumulation = river_accumulation(flow_accumulation, indices)
    geomorphic_flood_index = np.zeros(hand.shape, dtype=float32)

    for i in range(0, len(hand), 1):
        for j in range(0, len(hand[0]), 1):
            if hand[i, j] == -100:
                geomorphic_flood_index[i, j] = -100
            else:
                geomorphic_flood_index[i, j] = np.log(
                    scale_factor *
                    (np.power(river_flow_accumulation[i, j], expoent)) /
                    (hand[i, j] + 0.01))

    return geomorphic_flood_index


@jit
def ln_hl_H_sequential_jit(hand, flow_accumulation, expoent, scale_factor):

    ln_hl_H = np.zeros(hand.shape, dtype=float32)

    for i in range(0, len(hand), 1):
        for j in range(0, len(hand[0]), 1):
            if hand[i, j] == -100:
                ln_hl_H[i, j] = -100
            elif hand[i, j] == 0:
                ln_hl_H[i, j] = 0
            else:
                if flow_accumulation[i, j] == 0:
                    ln_hl_H[i,
                            j] = np.log(scale_factor * (np.power(1, expoent)) /
                                        (hand[i, j] + 0.01))
                else:
                    ln_hl_H[i, j] = np.log(
                        scale_factor *
                        (np.power(flow_accumulation[i, j], expoent)) /
                        (hand[i, j] + 0.01))

    return ln_hl_H


def ln_hl_H_sequential(hand, flow_accumulation, expoent, scale_factor):
    '''
    Sequential method for the ln(hl/HAND) index

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    flow_accumulation : int
        Flow accumulation. Represent the upslope area as source of hazard.
    expoent : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.

    Returns
    -------
    ln_hl_H : float array
        ln(hl/H) index array.

    '''
    ln_hl_H = np.where(
        hand == -100, -100,
        np.where(
            hand == 0, 0,
            np.log(scale_factor * (np.power(
                np.where(flow_accumulation == 0, 1, flow_accumulation),
                expoent)) / hand)))
    return ln_hl_H


@jit
def river_accumulation(flow_accumulation, indices):
    '''
    Method that return the array with the river cell flow accumulation.

    Parameters
    ----------
    flow_accumulation : int
        Flow accumulation. 
    indices : indices : int array
        River cell index.

    Returns
    -------
    river_flow_accumulation : int
        Flow accumulation of the river cell. Represent the river as source of hazard.

    '''
    row, col = flow_accumulation.shape
    flow_accumulation = np.asarray(flow_accumulation).reshape(-1)
    indices = np.asarray(indices).reshape(-1)
    river_flow_accumulation = np.zeros(row * col, float32)

    river_flow_accumulation = np.where(indices != -100,
                                       flow_accumulation[indices],
                                       flow_accumulation[0])

    river_flow_accumulation = river_flow_accumulation.reshape(row, col)

    return river_flow_accumulation


def gfi_calculator(hand, flow_accumulation, indices, n_gfi, scale_factor,
                   size):
    '''
    Method responsible for the partioning of the matrix

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    flow_accumulation : int
        Flow accumulation. Represent the river as source of hazard.
    indices : indices : int array
        River cell index.
    n_gfi : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.

    Returns
    -------
    gfi : float array
        geomorphic flood index array.

    '''
    row_size = len(hand)
    col_size = len(hand[0])

    div_col = 2
    div_row = 2

    boundary_row, boundary_column = divisor(row_size, col_size, div_row,
                                            div_col)

    flow_accumulation = river_accumulation(flow_accumulation, indices)

    gfi = np.zeros((row_size, col_size))

    boundary_row = np.insert(boundary_row, div_row, row_size)
    boundary_row = np.insert(boundary_row, 0, 0)
    boundary_column = np.insert(boundary_column, div_col, col_size)
    boundary_column = np.insert(boundary_column, 0, 0)

    for m in range(0, div_row + 1, 1):
        for n in range(0, div_col + 1, 1):

            mS = boundary_row[m]
            mE = boundary_row[m + 1]
            nS = boundary_column[n]
            nE = boundary_column[n + 1]

            gfi[mS:mE, nS:nE] = geomorphic_flood_index_cpu(
                hand[mS:mE, nS:nE], flow_accumulation[mS:mE, nS:nE], n_gfi,
                scale_factor, size)

    return gfi


def geomorphic_flood_index_cpu(hand,
                               river_flow_accumulation,
                               expoent,
                               scale_factor,
                               size,
                               blocks=0,
                               threads=0):
    '''
    Method responsible for the host/device data transfer

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    river_flow_accumulation : int
        Flow accumulation. Represent the river as source of hazard.
    expoent : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.
    blocks : int, optional
        Number of block of threads. The default is 0.
    threads : int, optional
        number of threads in each block. The default is 0.

    Returns
    -------
    gfi : float array
        geomorphic flood index array.

    '''
    row = len(hand)
    col = len(hand[0])
    if blocks == 0 and threads == 0:
        threads = 256
        blocks = math.ceil((row * col) / threads)

    hand = np.asarray(hand).reshape(-1)
    river_flow_accumulation = np.asarray(river_flow_accumulation).reshape(-1)

    gfi = np.zeros((row * col), dtype='float32')

    river_flow_accumulation = cuda.to_device(river_flow_accumulation)
    hand = cuda.to_device(hand)

    gfi = cuda.to_device(gfi)

    geomorphic_flood_index_gpu[blocks,
                               threads](hand, river_flow_accumulation, gfi,
                                        expoent, scale_factor, size)

    gfi = gfi.copy_to_host()
    gfi = gfi.reshape(row, col)

    return gfi


@cuda.jit
def geomorphic_flood_index_gpu(hand, river_flow_accumulation, gfi, expoent,
                               scale_factor, size):
    '''
    GPU GFI index method

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    river_flow_accumulation : int
        Flow accumulation. Represent the river as source of hazard.
    gfi : float array
        geomorphic flood index array. Initialized as zeros.
    expoent : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.

    '''
    i = cuda.grid(1)
    if i >= 0 and i < len(hand):
        if hand[i] <= -100:
            gfi[i] = -100
        else:
            gfi[i] = math.log(scale_factor * (math.pow(
                (river_flow_accumulation[i] * (size * size)), expoent)) /
                              (hand[i] + 0.01))


def ln_hl_H_calculator(hand, flow_accumulation, n_gfi, scale_factor, size):
    '''
    Method responsible for the partioning of the matrix

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    fac : int
        Flow accumulation. 
    n_gfi : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.

    Returns
    -------
    ln_hl_H : float array
        ln(hl/H) index array.

    '''
    row_size = len(hand)
    col_size = len(hand[0])

    div_col = 0
    div_row = 0

    bRow, bCol = divisor(row_size, col_size, div_row, div_col)

    lnhlh = np.zeros((row_size, col_size))
    bRow = np.insert(bRow, div_row, row_size)
    bRow = np.insert(bRow, 0, 0)
    bCol = np.insert(bCol, div_col, col_size)
    bCol = np.insert(bCol, 0, 0)

    for m in range(0, div_row + 1, 1):
        for n in range(0, div_col + 1, 1):
            mS = bRow[m]
            mE = bRow[m + 1]
            nS = bCol[n]
            nE = bCol[n + 1]

            lnhlh[mS:mE, nS:nE] = ln_hl_H_cpu(hand[mS:mE, nS:nE],
                                              flow_accumulation[mS:mE, nS:nE],
                                              n_gfi, scale_factor, size)

    return lnhlh


def ln_hl_H_cpu(hand,
                flow_accumulation,
                expoent,
                scale_factor,
                size,
                blocks=0,
                threads=0):
    '''
    Method responsible for the host/device data transfer

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    flow_accumulation : int
        Flow accumulation. 
    expoent : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.
    blocks : int, optional
        Number of block of threads. The default is 0.
    threads : int, optional
        number of threads in each block. The default is 0.

    Returns
    -------
    ln_hl_H : float array
        ln(hl/H) index array.

    '''
    row = len(hand)
    col = len(hand[0])
    if blocks == 0 and threads == 0:
        threads = 256
        blocks = math.ceil((row * col) / threads)

    hand = np.asarray(hand).reshape(-1)
    flow_accumulation = np.asarray(flow_accumulation).reshape(-1)
    lnhlh = np.zeros((row * col), dtype='float32')

    flow_accumulation = cuda.to_device(flow_accumulation)
    hand = cuda.to_device(hand)
    lnhlh = cuda.to_device(lnhlh)

    ln_hl_H_gpu[blocks, threads](hand, flow_accumulation, lnhlh, expoent,
                                 scale_factor, size)

    lnhlh = lnhlh.copy_to_host()
    lnhlh = lnhlh.reshape(row, col)

    return lnhlh


@cuda.jit
def ln_hl_H_gpu(hand, flow_accumulation, ln_hl_H, expoent, scale_factor, size):
    '''
    GPU ln(hl/HAND) index method

    Parameters
    ----------
    hand : int or flat array
        Height above the nearest drainage.
    flow_accumulation : int
        Flow accumulation. Represent the river as source of hazard.
    ln_hl_H : float array
        lnhlh index array. Initialized as zeros.
    expoent : float
        Expoent (<1) Calibrated for the region.
    scale_factor : float
        Scale factor.

    Returns 
    --------
    ln_hl_H : float array
        ln(hl/H) index array.

    '''
    i = cuda.grid(1)
    if i >= 0 and i < len(hand):
        if hand[i] <= -100:
            ln_hl_H[i] = -100
        else:
            if flow_accumulation[i] == 0:
                ln_hl_H[i] = math.log(
                    (scale_factor * math.pow(1 * (size * size), expoent)) /
                    (hand[i] + 0.01))
            else:
                ln_hl_H[i] = math.log(
                    (scale_factor *
                     math.pow(flow_accumulation[i] *
                              (size * size), expoent)) / (hand[i] + 0.01))



import numpy as np
import math


def divisor(row_length, column_length, row_division, column_division):
    boundary_column = np.array([], dtype=int)
    boundary_row = np.array([], dtype=int)

    for i in range(0, row_division, 1):
        boundary_row = np.append(
            boundary_row,
            [math.floor((i + 1) * row_length / (row_division + 1))])
    for i in range(0, column_division, 1):
        boundary_column = np.append(
            boundary_column,
            [math.floor((i + 1) * column_length / (column_division + 1))])

    return boundary_row, boundary_column


import numpy as np
import math


def minMaxScale(mat, mn, mx, nodata):
    scaled = np.where(mat == nodata, np.nan, mat)
    scaled = np.where(np.isnan(mat), scaled, (scaled - mn) / (mx - mn))

    return scaled


def calibration(descriptor_matrix, comparison_matrix, under):
    '''
    Return the best threshold value (th) for the linear binary classification.
    It is found by iteratively changing the value.

    Parameters
    ----------
    descriptor_matrix : int or float array
        Descriptor matrix
    comparison_matrix : int
        Benchmark flood map matrix
    under : str
        Direction which the threshold classification is applied.

    Returns
    -------
    threshold : float
        Threshold value that returns the best fit index.

    '''
    correctness_index_1, fit_index_1, _ = avaliacao(
        binary_map(descriptor_matrix, 25 / 100, under), comparison_matrix)
    correctness_index_2, fit_index_2, _ = avaliacao(
        binary_map(descriptor_matrix, 50 / 100, under), comparison_matrix)
    correctness_index_3, fit_index_3, _ = avaliacao(
        binary_map(descriptor_matrix, 75 / 100, under), comparison_matrix)

    if fit_index_3 > fit_index_2:
        if fit_index_3 > fit_index_1:
            fit_index = fit_index_3
            iteration_value = 75
        else:
            iteration_value = 25
            fit_index = fit_index_1
    else:
        if fit_index_2 > fit_index_1:
            fit_index = fit_index_2
            iteration_value = 50
        else:
            iteration_value = 25
            fit_index = fit_index_1

    for i in range(iteration_value - 20, iteration_value + 30, 10):
        correctness_index_1, iteration_fit_value, _ = avaliacao(
            binary_map(descriptor_matrix, i / 100, under), comparison_matrix)
        if iteration_fit_value >= fit_index:
            fit_index = iteration_fit_value
            threshold = i

    iteration_value = threshold
    for i in range(iteration_value - 5, iteration_value + 6, 1):
        correctness_index_1, iteration_fit_value, _ = avaliacao(
            binary_map(descriptor_matrix, i / 100, under), comparison_matrix)
        if iteration_fit_value > fit_index:
            fit_index = iteration_fit_value
            threshold = i

    iteration_value = threshold * 10
    threshold = iteration_value
    for i in range(iteration_value - 10, iteration_value + 11, 1):
        correctness_index_1, iteration_fit_value, _ = avaliacao(
            binary_map(descriptor_matrix, i / 1000, under), comparison_matrix)
        if iteration_fit_value > fit_index:
            fit_index = iteration_fit_value
            threshold = i

    iteration_value = threshold * 10
    threshold = iteration_value
    for i in range(iteration_value - 10, iteration_value + 11, 1):
        correctness_index_1, iteration_fit_value, _ = avaliacao(
            binary_map(descriptor_matrix, i / 10000, under), comparison_matrix)
        if iteration_fit_value > fit_index:
            fit_index = iteration_fit_value
            threshold = i

    return threshold / 10000


def binary_map(descriptor_matrix, threshold, under):
    '''
    Generates the linear binary map matrix from the descriptor matrix and
    threshold value

    Parameters
    ----------
    descriptor_matrix : int or float
        Terrain descriptor matrix.
    threshold : float
        Limit value used to classify the descriptors cells
    under : str
        Direction which the threshold classification is applied.

    Returns
    -------
    descriptor_binary : int8 array
        Binary flood map generated from the descriptor map.
        0 = not flooded cell, 1 = flooded cell

    '''
    descriptor_matrix = np.where(descriptor_matrix == descriptor_matrix[0, 0],
                                 np.nan, descriptor_matrix)

    if under == 'under':
        descriptor_binary = np.where(
            np.isnan(descriptor_matrix), 0,
            np.where(descriptor_matrix <= threshold, 1, 0))
    else:
        descriptor_binary = np.where(
            np.isnan(descriptor_matrix), 0,
            np.where(descriptor_matrix >= threshold, 1, 0))

    return descriptor_binary


def avaliacao(descriptor_flood_map, comparison_flood_map):
    '''
    Calculates the performance indexes: Fit index and Correctness index
    Also returns the resulting linear binary classified matrix

    Parameters
    ----------
    descriptor_flood_map : int8 array
        Binary flood map generated from the terrain descriptor.
    comparison_flood_map : int8 array
        Binary flood map generated from the benchmark flood map

    Returns
    -------
    correctness_index : float
        Correctness index.
    fit_index : float
        Fit index.
    result : int8 array
        Linear binary matrix, where 0 = true negative, 1 = false positive
        2 = false negative and 3 = true positive

    '''
    comparison_flood_map[comparison_flood_map == 1] = 2
    comparison_flood_map[comparison_flood_map == -100] = 0

    result = descriptor_flood_map + comparison_flood_map
    elements, count = np.unique(result, return_counts=True)

    if not np.any(elements == 0):
        count = np.insert(count, 0, 0)
        elements = np.insert(elements, 0, 0)
    if not np.any(elements == 1):
        count = np.insert(count, 1, 0)
        elements = np.insert(elements, 1, 1)
    if not np.any(elements == 2):
        count = np.insert(count, 2, 0)
        elements = np.insert(elements, 2, 2)
    if not np.any(elements == 3):
        count = np.insert(count, 3, 0)
        elements = np.insert(elements, 3, 3)

    correctness_index = correctness(count)
    fit_index = fit(count)

    return correctness_index, fit_index, result


def correctness(count):
    '''
    Correctness index:
        [true positive]/([false negatives] + [true positives])
    Represents the percentage of flood cells that were correctly classified.

    Parameters
    ----------
    count : int array
        Array with true/false positives/negatives.

    Returns
    -------
    correctness: float
        Correctness index.

    '''
    return ((count[3]) / (count[2] + count[3]))


def fit(count):
    '''  
    Fit index:
        [true positive]/([false negatives] + [false positives] + [true positives])
    Represents the percentage of cells that were correctly classified.

    Parameters
    ----------
    count : int array
        Array with true/false positives/negatives.

    Returns
    -------
    fit: float
        Fit index.

    '''
    return ((count[3]) / (count[3] + count[2] + count[1]))



from numba import cuda, jit
import numpy as np
import math

from descriptools.helpers import divisor


@jit
def slope_sequential_jit(dem, px):
    '''
    Return the highest slope to a neighbouring cell. Sequential implementation
    
    Parameters:
        dem : Digital evelation model
        px : Raster pixel dimensions
    '''
    row, col = dem.shape
    slope = np.zeros((row, col))
    for i in range(0, row, 1):
        for j in range(0, col, 1):
            aux = 0

            if dem[i, j] == -100:
                slope[i, j] = -100
                continue

            for y in range(-1, 2, 1):
                for x in range(-1, 2, 1):
                    if i + y < 0 or i + y >= row or j + x < 0 or j + x >= col:
                        continue
                    if x == 0 and y == 0:
                        continue
                    if dem[i + y, j + x] == -100:
                        continue

                    if x == 0 or y == 0:
                        if aux < (dem[i, j] - dem[i + y, j + x]) / px:
                            aux = (dem[i, j] - dem[i + y, j + x]) / px
                        else:
                            continue
                    else:
                        if aux < (dem[i, j] - dem[i + y, j + x]) / (px *
                                                                     math.sqrt(2.0)):
                            aux = (dem[i, j] - dem[i + y, j + x]) / (px *
                                                                      math.sqrt(2.0))
                        else:
                            continue
            slope[i, j] = aux * 100
    return slope


def slope_sequential(dem, px):
    '''
    Return the highest slope to a neighbouring cell. Sequential implementation
    
    Parameters:
        dem : Digital evelation model
        px : Raster pixel dimensions
    '''
    row, col = dem.shape
    slope = np.zeros((row, col))
    for i in range(2, row, 1):
        for j in range(190, col, 1):
            aux = 0

            if dem[i, j] == -100:
                slope[i, j] = -100
                continue

            for y in range(-1, 2, 1):
                for x in range(-1, 2, 1):
                    if i + y < 0 or i + y >= row or j + x < 0 or j + x >= col:
                        continue
                    if x == 0 and y == 0:
                        continue
                    if dem[i + y, j + x] == -100:
                        continue

                    if x == 0 or y == 0:
                        if aux < (dem[i, j] - dem[i + y, j + x]) / px:
                            aux = (dem[i, j] - dem[i + y, j + x]) / px
                        else:
                            continue
                    else:
                        # var = (dem[i,j] - dem[i+y,j+x])/(px*1.4142)
                        if aux < (dem[i, j] - dem[i + y, j + x]) / (px *
                                                                    math.sqrt(2.0)):
                            aux = (dem[i, j] - dem[i + y, j + x]) / (px *
                                                                      math.sqrt(2.0))
                        else:
                            continue
            slope[i, j] = aux
    return slope


def sloper(dem, px):
    '''
    Method responsible for the matrix dimension division

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    px : int or float
        Raster pixel dimension.

    Returns
    -------
    slope : float
        Highest slope of the neighbouring cells.

    '''
    #Essa é a função responsável pela divisão
    #Slope é simples, divide em x partes e adiciona colunas dos lados
    #Se pegar coluna/linha de fora, adiciona como nan, se não pega da outra divisão

    div_col = 0
    div_row = 0

    row_size = len(dem)
    col_size = len(dem[0])

    bRow, bCol = divisor(row_size, col_size, div_row, div_col)

    slope = np.zeros((row_size, col_size))

    bRow = np.insert(bRow, div_row, row_size)
    bRow = np.insert(bRow, 0, 0)
    bCol = np.insert(bCol, div_col, col_size)
    bCol = np.insert(bCol, 0, 0)

    for m in range(0, div_row + 1, 1):
        for n in range(0, div_col + 1, +1):

            mS = bRow[m]
            mE = bRow[m + 1]
            nS = bCol[n]
            nE = bCol[n + 1]

            # print(dem[mS:mE,nS:nE])
            extra = np.array([0, 0, 0, 0])

            if m == 0:
                extra[0] = 1  #cima
            if n == 0:
                extra[1] = 1  #esquerda
            if n == div_col:
                extra[2] = 1  #direita
            if m == div_row:
                extra[3] = 1  #baixo

            slope[mS:mE, nS:nE] = slope_cpu(
                dem[mS - 1 + extra[0]:mE + 1 - extra[3],
                    nS - 1 + extra[1]:nE + 1 - extra[2]], px, extra)

    return slope


def slope_cpu(dem, px, extra, blocks=0, threads=0):
    '''
    Method responsible for the host-device data transfer and method calling.

    Parameters
    ----------
    dem : int or float
        digital elevation model.
    px : int or float
        raster pixel dimension.
    extra : int or float
        Extra columns and rows necessary for the first/last column/row cells.
    blocks : int, optional
        Number of block of threads. The default is 0.
    threads : int, optional
        number of threads in each block. The default is 0.

    Returns
    -------
    slope : int or float
        DESCRIPTION.

    '''
    if extra[0] == 1:
        dem = np.insert(dem, 0, -100, axis=0)
    if extra[1] == 1:
        dem = np.insert(dem, 0, -100, axis=1)
    if extra[2] == 1:
        dem = np.insert(dem, len(dem[0]), -100, axis=1)
    if extra[3] == 1:
        dem = np.insert(dem, len(dem), -100, axis=0)

    row = len(dem)
    col = len(dem[0])

    if blocks == 0 and threads == 0:
        threads = 256
        blocks = math.ceil((row * col) / threads)
        # TODO: check again

    dem = np.asarray(dem).reshape(-1)
    slope = np.zeros((row * col), dtype='float32')

    dem = cuda.to_device(dem)
    slope = cuda.to_device(slope)

    slope_gpu[blocks, threads](dem, slope, px, row, col)

    slope = slope.copy_to_host()
    slope = slope.reshape(row, col)
    slope = np.delete(slope, 0, axis=0)
    slope = np.delete(slope, 0, axis=1)
    slope = np.delete(slope, len(slope) - 1, axis=0)
    slope = np.delete(slope, len(slope[0]) - 1, axis=1)
    return slope


@cuda.jit
def slope_gpu(dem, slope, px, row, column):
    '''
    GPU slope method

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    slope : float
        Slope matrix that will be filled and returned. Initialized as zeros.
    px : int or float
        Rasters pixel dimension size.
    row : int
        Number of rows of the 2-D matrix.
    column : int
        Number of columns of the 2-D matrix.

    '''
    i = cuda.grid(1)

    if i >= 0 and i < column * row:
        if dem[i] <= -100:
            slope[i] = -100
        else:
            if i % column == 0:
                slope[i] = -100
            elif i < column:
                slope[i] = -100
            elif i % column == column - 1:
                slope[i] = -100
            elif i >= column * (row - 1):
                slope[i] = -100
            else:
                aux = 0
                for y in range(-1, 2, 1):
                    for x in range(-1, 2, 1):
                        pos = i + (y * column) + x
                        if dem[pos] == -100:
                            continue
                        if x == 0 or y == 0:
                            if aux < (dem[i] - dem[pos]) / px:
                                aux = (dem[i] - dem[pos]) / px
                            else:
                                continue
                        else:
                            if aux < (dem[i] - dem[pos]) / (px * math.sqrt(2.0)):
                                aux = (dem[i] - dem[pos]) / (px * math.sqrt(2.0))
                            else:
                                continue
            slope[i] = aux * 100



from numba import cuda, jit, int32, float32
import numpy as np
import math

from descriptools.helpers import divisor


def flow_distance_indexes_sequential(flow_direction,
                                     river_matrix,
                                     px,
                                     flow_distance=np.array([])):
    '''
    Flow distance to the nearest drainage sequential method. Also preemptively 
    calculates the row and columns required for the matrix division.

    Parameters
    ----------
    flow_direction : int8 array
        Flow direction.
    river_matrix : int8 array
        Binary matrix that 1- river cell 0-not river cell.
    px : int or float
        Raster pixel dimension.
    flow_distance : flow distance float array, optional
        Flow distance. The default is an empty array.
        This input is used for the boundary array calculation for matrix subdivision.

    Returns
    -------
    flow_distance : float array
        Flow distance.
    indices : int array
        River cell index.

    '''
    new = 0
    if flow_distance.size == 0:
        flow_distance = np.zeros(
            [len(flow_direction), len(flow_direction[0])], dtype='float32')
        new = 1

    indices = np.zeros(
        [len(flow_direction), len(flow_direction[0])], dtype='int')
    for i in range(0, len(flow_direction), 1):
        for j in range(0, len(flow_direction[0]), 1):
            if flow_direction[i, j] == 0:
                flow_distance[i, j] = -100
                indices[i, j] = -100
            elif new == 0 and flow_distance[i, j] != -50:
                continue
            else:
                y = i
                x = j
                dist = 0
                loop = 0
                isnan = 0
                while (river_matrix[y, x] != 1):
                    if y == 0 and (flow_direction[y, x] == 32
                                   or flow_direction[y, x] == 64
                                   or flow_direction[y, x] == 128):
                        isnan = 1
                        break
                    elif y == len(flow_distance) - 1 and (
                            flow_direction[y, x] == 2 or flow_direction[y, x]
                            == 4 or flow_direction[y, x] == 8):
                        isnan = 1
                        break
                    elif x == 0 and (flow_direction[y, x] == 32
                                     or flow_direction[y, x] == 16
                                     or flow_direction[y, x] == 8):
                        isnan = 1
                        break
                    elif x == len(flow_distance[0]) - 1 and (
                            flow_direction[y, x] == 128 or flow_direction[y, x]
                            == 1 or flow_direction[y, x] == 2):
                        isnan = 1
                        break

                    if flow_direction[y, x] == 1:
                        x += 1
                        dist += px
                    elif flow_direction[y, x] == 2:
                        x += 1
                        y += 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == 4:
                        y += 1
                        dist += px
                    elif flow_direction[y, x] == 8:
                        x -= 1
                        y += 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif flow_direction[y, x] == 16:
                        x -= 1
                        dist += px
                    elif flow_direction[y, x] == 32:
                        x -= 1
                        y -= 1
                        dist += px * math.sqrt(2.0) # FIX: Square root
                    elif flow_direction[y, x] == 64:
                        y -= 1
                        dist += px
                    elif flow_direction[y, x] == 128:
                        x += 1
                        y -= 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root

                    elif flow_direction[y, x] == 255:
                        isnan = 1
                        break

                    loop += 1
                    if loop >= 5000:
                        isnan = 1
                        break

                if isnan == 1:
                    indices[i, j] = -100
                    flow_distance[i, j] = -100
                else:
                    flow_distance[i, j] = dist
                    indices[i, j] = y * len(flow_direction[0]) + x

    return flow_distance, indices


@jit
def fdist_indexes_sequential_jit(fdr,
                                 river,
                                 px,
                                 fdist=np.array([[], []], 'float32')):
    '''
    Flow distance to the nearest drainage sequential method. Also preemptively 
    calculates the row and columns required for the matrix division.

    Parameters
    ----------
    fdr : int8 array
        Flow direction.
    river : int8 array
        Binary matrix that 1- river cell 0-not river cell.
    px : int or float
        Raster pixel dimension.
    fdist : flow distance float array, optional
        Flow distance. The default is an empty array.
        This input is used for the boundary array calculation for matrix subdivision.

    Returns
    -------
    fdist : float array
        Flow distance.
    indices : int array
        River cell index.

    '''
    new = 0
    if fdist.size == 0:
        fdist = np.zeros(fdr.shape, dtype=float32)
        new = 1

    indices = np.zeros(fdr.shape, int32)
    for i in range(0, len(fdr), 1):
        for j in range(0, len(fdr[0]), 1):
            if new == 0 and fdist[i, j] != -50:
                continue
            elif fdr[i, j] == 0:
                fdist[i, j] = -100
                indices[i, j] = -100
            else:
                y = i
                x = j
                dist = 0
                loop = 0
                isnan = 0
                while (river[y, x] != 1):
                    if y == 0 and (fdr[y, x] == 32 or fdr[y, x] == 64
                                   or fdr[y, x] == 128):
                        isnan = 1
                        break
                    elif y == len(fdist) - 1 and (fdr[y, x] == 2 or fdr[y, x]
                                                  == 4 or fdr[y, x] == 8):
                        isnan = 1
                        break
                    elif x == 0 and (fdr[y, x] == 32 or fdr[y, x] == 16
                                     or fdr[y, x] == 8):
                        isnan = 1
                        break
                    elif x == len(fdist[0]) - 1 and (fdr[y, x] == 128
                                                     or fdr[y, x] == 1
                                                     or fdr[y, x] == 2):
                        isnan = 1
                        break

                    if fdr[y, x] == 1:
                        x += 1
                        dist += px
                    elif fdr[y, x] == 2:
                        x += 1
                        y += 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif fdr[y, x] == 4:
                        y += 1
                        dist += px
                    elif fdr[y, x] == 8:
                        x -= 1
                        y += 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif fdr[y, x] == 16:
                        x -= 1
                        dist += px
                    elif fdr[y, x] == 32:
                        x -= 1
                        y -= 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root
                    elif fdr[y, x] == 64:
                        y -= 1
                        dist += px
                    elif fdr[y, x] == 128:
                        x += 1
                        y -= 1
                        dist += px * math.sqrt(2.0)  # FIX: Square root

                    elif fdr[y, x] == 0:
                        isnan = 1
                        break

                    loop += 1
                    if loop >= 5000:
                        isnan = 1
                        break

                if isnan == 1:
                    indices[i, j] = -100
                    fdist[i, j] = -100
                else:
                    fdist[i, j] = dist
                    indices[i, j] = y * len(fdr[0]) + x

    return fdist, indices


def flow_hand_index(dem_raster,
                    flow_direction_matrix,
                    river_matrix,
                    px,
                    division_column=0,
                    division_row=0):
    '''
    Overall method responsible for the matrix partitioning and
    defining dimensions.

    Parameters
    ----------
    dem_raster : int or float array
        Digital elevation model raster.
    flow_direction_matrix : int8 array
        Flow direction.
    river_matrix : int8 array
        Binary matrix that 1- river cell 0-not river cell.
    px : int or float
        Raster pixel dimension.

    Returns
    -------
    fdist : float array
        Flow distance.
    indices : int array
        River cell index.
    hand : int or flat array
        Height above the nearest drainage.

    '''
    row = len(dem_raster)
    col = len(dem_raster[0])

    boundary_row, boundary_column = divisor(row, col, division_row,
                                            division_column)

    flow_distance = np.zeros((row, col),dtype='float32')
    indices = np.zeros((row, col), dtype='int')

    if division_row > 0 or division_column > 0:
        flow_distance[:, boundary_column] = -50
        flow_distance[boundary_row, :] = -50
        flow_distance, indices = fdist_indexes_sequential_jit(
            flow_direction_matrix, river_matrix, px, flow_distance)

    boundary_row = np.insert(boundary_row, division_row, row)
    boundary_row = np.insert(boundary_row, 0, -1)
    boundary_column = np.insert(boundary_column, division_column, col)
    boundary_column = np.insert(boundary_column, 0, -1)

    part = 0

    for m in range(0, division_row + 1, 1):
        for n in range(0, division_column + 1, 1):
            out = np.zeros((4))
            distance_up = np.zeros([1])
            distance_left = np.zeros([1])
            distance_right = np.zeros([1])
            distance_down = np.zeros([1])
            
            index_up    = np.zeros([1])
            index_left  = np.zeros([1])
            index_right = np.zeros([1])
            index_down  = np.zeros([1])

            mS = boundary_row[m]
            mE = boundary_row[m + 1]
            nS = boundary_column[n]
            nE = boundary_column[n + 1]

            if division_row > 0:
                if part < (division_column + 1) * division_row:
                    out[3] = 1
                    distance_down = flow_distance[mE, nS + 1:nE]
                    index_down = indices[mE, nS + 1:nE]
                    if n != division_column:
                        distance_down = np.insert(distance_down, nE - (nS + 1),
                                            flow_distance[mE, nE])
                        index_down = np.insert(index_down, nE - (nS + 1),
                                            indices[mE, nE])
                        
                    if n != 0:
                        distance_down = np.insert(distance_down, 0, flow_distance[mE, nS])
                        index_down = np.insert(index_down, 0, indices[mE, nS])

                if part > division_column:
                    out[0] = 1
                        
                    distance_up = flow_distance[mS, nS + 1:nE]
                    index_up = indices[mS, nS + 1:nE]
                    
                    if n != division_column:
                        distance_up = np.insert(distance_up, nE - (nS + 1),
                                            flow_distance[mS, nE])
                        index_up = np.insert(index_up, nE - (nS + 1),
                                            indices[mS, nE])
                    if n != 0:
                        distance_up = np.insert(distance_up, 0, flow_distance[mS, nS])
                        index_up = np.insert(index_up, 0, indices[mS, nS])

            if division_column > 0:
                if part % (division_column + 1) != 0:
                    out[1] = 1
                    distance_left = flow_distance[mS + 1:mE, nS]
                    index_left = indices[mS + 1:mE, nS]
                    
                    if m != division_row:
                        distance_left = np.insert(distance_left, mE - (mS + 1),
                                            flow_distance[mE, nS])
                        index_left = np.insert(index_left, mE - (mS + 1),
                                            indices[mE, nS])
                        
                    if m != 0:
                        distance_left = np.insert(distance_left, 0, flow_distance[mS, nS])
                        index_left = np.insert(index_left, 0, indices[mS, nS])

                if part % (division_column + 1) != division_column:
                    out[2] = 1
                    distance_right = flow_distance[mS + 1:mE, nE]
                    index_right = indices[mS + 1:mE, nE]
                    
                    if m != division_row:
                        distance_right = np.insert(distance_right, mE - (mS + 1),
                                            flow_distance[mE, nE])
                        index_right = np.insert(index_right, mE - (mS + 1),
                                            indices[mE, nE])
                        
                    if m != 0:
                        distance_right = np.insert(distance_right, 0, flow_distance[mS, nE])
                        index_right = np.insert(index_right, 0, indices[mS, nE])
                        

            size = max(len(distance_up), len(distance_left), len(distance_right), len(distance_down))

            bound = np.zeros((4, size))
            bound_index = np.zeros((4, size))
            
            bound[0, 0:len(distance_up)] = distance_up
            bound_index[0, 0:len(index_up)] = index_up
            
            bound[1, 0:len(distance_left)] = distance_left
            bound_index[1, 0:len(index_left)] = index_left
            
            bound[2, 0:len(distance_right)] = distance_right
            bound_index[2, 0:len(index_right)] = index_right
            
            bound[3, 0:len(distance_down)] = distance_down
            bound_index[3, 0:len(index_down)] = index_down       

            mS += 1
            nS += 1
            part += 1
            
            flow_distance[mS:mE,
                          nS:nE], indices[mS:mE,
                                          nS:nE] = flow_distance_index_cpu(
                                              dem_raster[mS:mE, nS:nE],
                                              flow_direction_matrix[mS:mE,
                                                                    nS:nE],
                                              river_matrix[mS:mE, nS:nE], px,
                                              bound, bound_index, out, 
                                              mS, nS, col)

            # indices[mS:mE, nS:nE] = index_calculator(indices[mS:mE, nS:nE], mS,
            #                                           nS, col)

    hand = hand_calculator(dem_raster, indices)

    return flow_distance, indices, hand


def hand_calculator(dem, indices):
    '''
    Method that calculates the HAND index for each cell and its river cell

    Parameters
    ----------
    dem : int or float
        Digital elevation model.
    indices : int array
        River cell index.

    Returns
    -------
    hand : int or flat array
        Height above the nearest drainage.

    '''
    row, col = dem.shape
    dem = np.asarray(dem).reshape(-1)
    indices = np.asarray(indices).reshape(-1)
    hand = np.zeros(dem.shape)

    hand = np.where((dem != -100) & (indices != -100), dem - dem[indices],
                    -100)
    hand = np.where((hand < 0) & (hand != -100), 0, hand)

    hand = hand.reshape(row, col)

    return hand


def index_calculator(river_indices, row_start, column_start, column_size):
    '''
    Method that transform the river index from the sub-matrix to the whole matrix

    Parameters
    ----------
    river_indices : int array
        Sub matrix river cell index.
    row_start : int
        Position of the first row of the sub-matrix.
    column_start : int
        Position of the first column of the sub-matrix.
    column_size : int
        Number of columns in the whole matrix.

    Returns
    -------
    river_indices : int array
        River cell index.

    '''
    row, col = river_indices.shape

    river_indices = np.where(
        river_indices == -100, -100,
        (np.floor(river_indices / col) + row_start) * column_size +
        river_indices % col + column_start)

    return river_indices


def flow_distance_index_cpu(dem,
                            flow_direction,
                            river_matrix,
                            px,
                            boundary_distance,
                            boundary_index,
                            out,
                            row_start,
                            col_start,
                            matrix_columns,
                            blocks=0,
                            threads=0):
    '''
    Method responsible for the host/device data transfer

    Parameters
    ----------
    dem : int or float array
        Digital elevation model raster.
    flow_direction : int8 array
        Flow direction.
    river_matrix : int8 array
        Binary matrix that 1- river cell 0-not river cell.
    px : int or float
        Raster pixel dimension.
    boundary_distance : float array
        Bounduary array that contains values of flow distance. 
        Necessary for flow directions that lead to outside of the sub-matrix domain.
    boundary_index : int array
        Bounduary array that contains values of river indexes. 
        Necessary for flow directions that lead to outside of the sub-matrix domain.
    out : int array
        Array that represents wich directions have another sub-matrix.
        1-True 0-False. Up, left, right, down.
    row_start : int
        Row index of the first value
    col_start : int
        Column index of the first value
    matrix_columns : int
        Number of columns of the whole matrix
    blocks : int, optional
        Number of block of threads. The default is 0.
    threads : int, optional
        number of threads in each block. The default is 0.

    Returns
    -------
    flow_distance : float array
        Flow distance.
    indices : int array
        River cell index.

    '''
    row = len(dem)
    col = len(dem[0])

    if blocks == 0 and threads == 0:
        threads = 256
        blocks = math.ceil((row * col) / threads)

    dem = np.asarray(dem).reshape(-1)
    flow_direction = np.asarray(flow_direction).reshape(-1)
    river_matrix = np.asarray(river_matrix).reshape(-1)

    flow_distance = np.zeros((row * col), dtype='float32')
    indices = np.zeros((row * col))

    flow_direction = cuda.to_device(flow_direction)
    flow_distance = cuda.to_device(flow_distance)
    indices = cuda.to_device(indices)
    out = cuda.to_device(out)
    boundary_distance = cuda.to_device(boundary_distance)
    boundary_index = cuda.to_device(boundary_index)

    flow_distance_index_gpu[blocks,
                            threads](flow_direction, river_matrix, px,
                                     flow_distance, indices, col, row,
                                     boundary_distance, boundary_index, out,
                                     row_start, col_start, matrix_columns)

    indices = indices.copy_to_host()
    flow_distance = flow_distance.copy_to_host()

    indices = indices.reshape(row, col)
    flow_distance = flow_distance.reshape(row, col)

    return flow_distance, indices


@cuda.jit
def flow_distance_index_gpu(flow_direction, river_matrix, px, flow_distance,
                            indices, col, row, boundary_distance,
                            boundary_index, out, row_start, col_start, matrix_columns):
    '''
    GPU flow distance and river cell indexes method

    Parameters
    ----------
    flow_direction : int8 array
        Flow direction.
    river_matrix : int8 array
        Binary matrix that 1- river cell 0-not river cell.
    px : int or float
        Raster pixel dimension.
    flow_distance : float array
        Flow distance.
    indices : int array
        River cell index.
    row : int
        Number of rows of the 2-D matrix.
    col : int
        Number of rows of the 2-D matrix.
    boundary_distance : float array
        Bounduary array that contains values of flow distance. 
        Necessary for flow directions that lead to outside of the sub-matrix domain.
    boundary_index : int array
        Bounduary array that contains values of river indexes. 
        Necessary for flow directions that lead to outside of the sub-matrix domain.
    out : int array
        Array that represents wich directions have another sub-matrix.
        1-True 0-False. row0 Up, row1 left, row2 right, row3 down.

    '''
    i = cuda.grid(1)
    if i >= 0 and i < col * row:
        if flow_direction[i] <= 0:
            flow_distance[i] = -100
            indices[i] = -100
        
        # elif i == 1 or i == 2:
        #     flow_distance[i] = -33
        #     indices[i] = -33
        else:
            if river_matrix[i] == 1:
                flow_distance[i] = 0
                # indices[i] = i
                indices[i] = (row_start + math.floor(i/col)) * matrix_columns + col_start + i%col
            else:
                pos = i
                isnan = 0
                dist = 0
                loop = 0
                loop1 = -10
                loop2 = -20
                loop3 = -30
                isout = 0
                while (river_matrix[pos] != 1):
                    if pos < col and (flow_direction[pos] == 32
                                      or flow_direction[pos] == 64
                                      or flow_direction[pos] == 128):
                        if out[0] == 0:
                            isnan = 1
                            break
                        else:
                            initial = pos
                            irow = int(initial)
                            if out[1] == 1:
                                irow += 1
                            if flow_direction[pos] == 32:
                                if pos == 0 and out[1] == 0:
                                    isnan = 1
                                    break
                                else:
                                    irow = irow - 1
                                    if boundary_distance[0][irow] == -100:
                                        isnan = 1
                                        break
                                    dist += (px * math.sqrt(2.0)
                                             ) + boundary_distance[0][irow]
                                    indices[i] = boundary_index[0][irow]
                                    isout = 1
                                    break
                            elif flow_direction[pos] == 128:
                                if pos == col - 1 and out[2] == 0:
                                    isnan = 1
                                    break
                                else:
                                    irow = irow + 1
                                    if boundary_distance[0][irow] == -100:
                                        isnan = 1
                                        break
                                    dist += px * math.sqrt(2.0) + boundary_distance[0][
                                        irow]
                                    indices[i] = boundary_index[0][irow]
                                    isout = 1
                                    break
                            else:
                                if boundary_distance[0][irow] == -100:
                                        isnan = 1
                                        break
                                dist += px + boundary_distance[0][irow]
                                indices[i] = boundary_index[0][irow]
                                isout = 1
                                break

                    elif pos % col == 0 and (flow_direction[pos] == 8
                                             or flow_direction[pos] == 16
                                             or flow_direction[pos] == 32):

                        if out[1] == 0:
                            isnan = 1
                            break
                        else:
                            initial = pos / col
                            irow = int(initial)
                            if out[0] == 1:
                                irow += 1
                            if flow_direction[pos] == 32:
                                irow = irow - 1
                                if  boundary_distance[1][irow] == -100:
                                    isnan = 1
                                    break
                                dist += px * math.sqrt(2.0) + boundary_distance[1][irow]
                                indices[i] = boundary_index[1][irow]
                                isout = 1
                                break
                            elif flow_direction[pos] == 8:
                                if out[3] == 0 and pos == row * (col - 1):
                                    isnan = 1
                                    break
                                else:
                                    irow = irow + 1
                                    if  boundary_distance[1][irow] == -100:
                                        isnan = 1
                                        break
                                    dist += px * math.sqrt(2.0) + boundary_distance[1][
                                        irow]
                                    indices[i] = boundary_index[1][irow]
                                    isout = 1
                                    break
                            else:
                                if  boundary_distance[1][irow] == -100:
                                    isnan = 1
                                    break
                                dist += px + boundary_distance[1][irow]
                                indices[i] = boundary_index[1][irow]
                                isout = 1
                                break

                    elif pos % col == (col -
                                       1) and (flow_direction[pos] == 128
                                               or flow_direction[pos] == 1
                                               or flow_direction[pos] == 2):
                        if out[2] == 0:
                            isnan = 1
                            break
                        else:
                            initial = (pos + 1) / col - 1
                            irow = int(initial)
                            if out[0] == 1:
                                irow += 1
                            if flow_direction[pos] == 128:
                                irow = irow - 1
                                if  boundary_distance[2][irow] == -100:
                                    isnan = 1
                                    break
                                dist += px * math.sqrt(2.0) + boundary_distance[2][irow]
                                indices[i] = boundary_index[2][irow]
                                isout = 1
                                break
                            elif flow_direction[pos] == 2:
                                if out[3] == 0 and pos == (row * col) - 1:
                                    isnan = 1
                                    break
                                else:
                                    irow = irow + 1
                                    if  boundary_distance[2][irow] == -100:
                                        isnan = 1
                                        break
                                    dist += px * math.sqrt(2.0) + boundary_distance[2][
                                        irow]
                                    indices[i] = boundary_index[2][irow]
                                    isout = 1
                                    break
                            else:
                                if boundary_distance[2][irow] == -100:
                                    isnan = 1
                                    break
                                dist += px + boundary_distance[2][irow]
                                indices[i] = boundary_index[2][irow]
                                isout = 1
                                break

                    elif pos >= (row - 1) * col and (
                            flow_direction[pos] == 2 or flow_direction[pos]
                            == 4 or flow_direction[pos] == 8):
                        if out[3] == 0:
                            isnan = 1
                            break
                        else:
                            initial = pos - (row - 1) * col
                            irow = int(initial)
                            if out[1] == 1:
                                irow += 1
                            if flow_direction[pos] == 8:
                                irow = irow - 1
                                if  boundary_distance[3][irow] == -100:
                                    isnan = 1
                                    break
                                dist += px * math.sqrt(2.0) + boundary_distance[3][irow]
                                indices[i] = boundary_index[3][irow]
                                isout = 1
                                break
                            elif flow_direction[pos] == 2:
                                irow += 1
                                if  boundary_distance[3][irow] == -100:
                                    isnan = 1
                                    break
                                dist += px * math.sqrt(2.0) + boundary_distance[3][irow]
                                indices[i] = boundary_index[3][irow]
                                isout = 1
                                break
                            else:
                                if  boundary_distance[3][irow] == -100:
                                    isnan = 1
                                    break
                                dist += px + boundary_distance[3][irow]
                                indices[i] = boundary_index[3][irow]
                                isout = 1
                                break

                    loop3 = loop2
                    loop2 = loop1
                    loop1 = pos

                    if flow_direction[pos] == 1:
                        pos += 1
                        dist += px
                    elif flow_direction[pos] == 2:
                        pos += 1 + col
                        dist += px * math.sqrt(2.0)
                    elif flow_direction[pos] == 4:
                        pos += col
                        dist += px
                    elif flow_direction[pos] == 8:
                        pos += col - 1
                        dist += px * math.sqrt(2.0)
                    elif flow_direction[pos] == 16:
                        pos += -1
                        dist += px
                    elif flow_direction[pos] == 32:
                        pos += -1 - col
                        dist += px * math.sqrt(2.0)
                    elif flow_direction[pos] == 64:
                        pos += -col
                        dist += px
                    elif flow_direction[pos] == 128:
                        pos += -col + 1
                        dist += px * math.sqrt(2.0)

                    if flow_direction[pos] == 0:
                        isnan = 1
                        break

                    if pos == loop1 or pos == loop2 or pos == loop3:
                        isnan = 1
                        break

                    loop += 1
                    if loop > 20000:
                        isnan = 1
                        break

                if isnan == 1:
                    flow_distance[i] = -100
                    indices[i] = -100
                else:
                    flow_distance[i] = dist
                    if isout == 0:
                        indices[i] = (row_start + math.floor(pos/col)) * matrix_columns + col_start + pos%col
                    # indices[i] = pos



from setuptools import setup

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md")) as f:
    long_description = f.read()

setup(
    name="descriptools",
    packages=["descriptools"],
    version="1.0.2",
    license="MIT",
    description=
    "A gpu-based toolbox for terrain descriptor calculation/delineation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    mantainers="JVBSouza, JHBoing",
    url="https://github.com/JVBSouza/descriptools",
    download_url="https://github.com/JVBSouza/descriptools/archive/master.zip",
    keywords=["hidrology", "gis", "flood"],
    install_requires=[
        "numpy",
        "numba",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Environment :: GPU :: NVIDIA CUDA"
    ],
)